from typing import Iterable, Optional, List

import torch
from rex.utils.logging import logger
from rex.utils.progress_bar import tqdm
from rex.data.label_encoder import LabelEncoder
from rex.data.transforms.base import TransformBase


def tagging_collate_fn(data):
    final_data = {
        "id": [],
        "input_ids": [],
        "mask": [],
        "labels": [],
    }
    for d in data:
        for key in final_data:
            final_data[key].append(d[key])

    final_data["token_ids"] = torch.tensor(final_data["token_ids"], dtype=torch.long)
    final_data["mask"] = torch.tensor(final_data["mask"], dtype=torch.long)
    if all(x is not None for x in final_data["labels"]):
        final_data["labels"] = torch.tensor(final_data["labels"], dtype=torch.long)
    else:
        final_data["labels"] = None

    return final_data


class CachedTaggingTransform(TransformBase):
    """Cached data transform for classification task."""

    def __init__(self, max_seq_len: int) -> None:
        super().__init__(max_seq_len)

        self.label_encoder = LabelEncoder()

    def transform(
        self,
        dataset: Iterable,
        desc: Optional[str] = "Transform",
        debug: Optional[bool] = False,
        **kwargs,
    ) -> List[dict]:
        final_data = []
        if debug:
            dataset = dataset[:500]
        transform_loader = tqdm(dataset, desc=desc)

        num_tot_ins = 0
        for data in transform_loader:
            text = data["text"]
            token_ids, mask = self.vocab.encode(
                list(text), max_seq_len=self.max_seq_len, update=True
            )
            labels = data["labels"]
            label_ids = self.label_encoder.update_encode_one(labels)
            label_ids = (
                label_ids[: len(token_ids)] + [-1] * self.max_seq_len - len(label_ids)
            )

            ins = {
                "id": data["id"],
                "input_ids": token_ids,
                "mask": mask,
                "labels": label_ids,
                "text": text,
            }
            final_data.append(ins)

        logger.info(transform_loader)
        logger.info(f"#Total Ins: {num_tot_ins}")

        return final_data

    def predict_transform(self, data: dict):
        """
        Args:
            data:
                {
                    "id": "blahblah",
                    "text": "text",
                }
        """
        text = data["text"]
        token_ids, mask = self.vocab.encode(
            list(text), max_seq_len=self.max_seq_len, update=False
        )

        obj = {
            "input_ids": token_ids,
            "mask": mask,
        }
        return obj



def calc_p_r_f1_from_tp_fp_fn(tp, fp, fn):
    try:
        p = tp / (tp + fp)
    except ZeroDivisionError:
        p = 0.0

    try:
        r = tp / (tp + fn)
    except ZeroDivisionError:
        r = 0.0

    try:
        f1 = 2 * p * r / (p + r)
    except ZeroDivisionError:
        f1 = 0.0

    return {"p": p, "r": r, "f1": f1, "tp": tp, "fp": fp, "fn": fn}



import torch
from torch.optim import Adam
from omegaconf import OmegaConf

from rex.utils.logging import logger
from rex.models.sent_pcnn import SentPCNN
from rex.models.bag_pcnn import PCNNOne
from rex.utils.io import load_line_json
from rex.utils.progress_bar import tqdm
from rex.utils.tensor_move import move_to_cuda_device
from rex.data.collate_fn import bag_re_collate_fn, re_collate_fn
from rex.data.transforms.sent_re import CachedMCMLSentRETransform
from rex.data.transforms.bag_re import CachedMCBagRETransform
from rex.data.manager import CachedManager
from rex.data.dataset import CachedBagREDataset, CachedDataset
from rex.tasks.base_task import TaskBase
from rex.metrics.classification import accuracy, mc_prf1, mcml_prf1


class MCMLSentRelationClassificationTask(TaskBase):
    def __init__(self, config: OmegaConf, **kwargs) -> None:
        super().__init__(config, **kwargs)

        self.transform = CachedMCMLSentRETransform(config.max_seq_len)
        self.data_manager = CachedManager(
            config.train_filepath,
            config.dev_filepath,
            config.test_filepath,
            CachedDataset,
            self.transform,
            load_line_json,
            config.train_batch_size,
            config.eval_batch_size,
            re_collate_fn,
            debug_mode=config.debug_mode,
            load_train_data=config.load_train_data,
            load_dev_data=config.load_dev_data,
            load_test_data=config.load_test_data,
        )

        self.model = SentPCNN(
            self.transform.vocab,
            self.transform.label_encoder.num_tags,
            config.dim_token_emb,
            config.max_seq_len,
            config.dim_pos_emb,
            config.num_filters,
            config.kernel_size,
            dropout=config.dropout,
        )
        self.model.to(self.config.device)
        self.optimizer = Adam(self.model.parameters(), lr=config.learning_rate)

    def train(self):
        for epoch_idx in range(self.config.num_epochs):
            logger.info(f"Epoch: {epoch_idx}/{self.config.num_epochs}")
            self.model.train()
            loader = tqdm(self.data_manager.train_loader, desc=f"Train(e{epoch_idx})")
            for batch in loader:
                del batch["id"]
                batch = move_to_cuda_device(batch, self.config.device)
                result = self.model(**batch)
                loss = result["loss"]
                loader.set_postfix({"loss": loss.item()})
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            logger.info(loader)

            measures = self.eval("dev")
            self.history["dev"].append(measures)
            test_measures = self.eval("test")
            self.history["test"].append(test_measures)

            is_best = False
            if measures["f1"] > self.best_metric:
                is_best = True
                self.best_metric = measures["f1"]
                self.best_epoch = epoch_idx
                self.no_climbing_cnt = 0
            else:
                self.no_climbing_cnt += 1

            if is_best and self.config.save_best_ckpt:
                self.save_ckpt(epoch=epoch_idx)

            logger.info(
                f"Epoch: {epoch_idx}, is_best: {is_best}, Dev: {measures}, Test: {test_measures}"
            )

            if (
                self.config.num_early_stop > 0
                and self.no_climbing_cnt > self.config.num_early_stop
            ):
                break

        logger.info(
            (
                f"Trial finished. Best Epoch: {self.best_epoch}, Dev: {self.history['dev'][self.best_epoch]}, "
                f"Best Test: {self.history['test'][self.best_epoch]}"
            )
        )

    @torch.no_grad()
    def eval(self, dataset_name):
        self.model.eval()
        name2loader = {
            "train": self.data_manager.train_loader,
            "dev": self.data_manager.dev_loader,
            "test": self.data_manager.test_loader,
        }
        loader = tqdm(name2loader[dataset_name], desc=f"{dataset_name} Eval")
        preds = []
        golds = []
        for batch in loader:
            batch = move_to_cuda_device(batch, self.config.device)
            golds.extend(batch["labels"].detach().cpu().tolist())
            del batch["id"]
            del batch["labels"]
            result = self.model(**batch)
            out = result["pred"]
            preds.extend(
                out.ge(self.config.pred_threshold).long().detach().cpu().tolist()
            )
        logger.info(loader)
        measures = mcml_prf1(preds, golds)
        return measures["macro"]

    @torch.no_grad()
    def predict(self, text, head, tail):
        self.model.eval()
        batch = self.transform.predict_transform(
            {"text": text, "head": head, "tail": tail}
        )
        tensor_batch = self.data_manager.collate_fn([batch])
        if "labels" in tensor_batch:
            del tensor_batch["labels"]
        if "id" in tensor_batch:
            del tensor_batch["id"]
        tensor_batch = move_to_cuda_device(tensor_batch, self.config.device)
        result = self.model(**tensor_batch)
        outs = result["pred"]
        outs = outs.ge(self.config.pred_threshold).long().detach().cpu().tolist()[0]
        preds = []
        for type_idx, pred in enumerate(outs):
            if pred == 1:
                preds.append(self.transform.label_encoder.id2label[type_idx])

        return preds


class MCMLBagRelationClassificationTask(TaskBase):
    def __init__(self, config: OmegaConf, **kwargs) -> None:
        super().__init__(config, **kwargs)

        self.transform = CachedMCBagRETransform(config.max_seq_len)
        self.data_manager = CachedManager(
            config.train_filepath,
            config.dev_filepath,
            config.test_filepath,
            CachedBagREDataset,
            self.transform,
            load_line_json,
            config.train_batch_size,
            config.eval_batch_size,
            bag_re_collate_fn,
            debug_mode=config.debug_mode,
            load_train_data=config.load_train_data,
            load_dev_data=config.load_dev_data,
            load_test_data=config.load_test_data,
        )

        self.model = PCNNOne(
            self.transform.vocab,
            self.transform.label_encoder.num_tags,
            config.dim_token_emb,
            config.max_seq_len,
            config.dim_pos_emb,
            config.num_filters,
            config.kernel_size,
            dropout=config.dropout,
        )
        self.model.to(self.config.device)
        self.optimizer = Adam(self.model.parameters(), lr=config.learning_rate)

        self.history = {"dev": [], "test": []}
        self.no_climbing_cnt = 0
        self.best_metric = -100.0
        self.best_epoch = -1

    def train(self):
        for epoch_idx in range(self.config.num_epochs):
            logger.info(f"Epoch: {epoch_idx}/{self.config.num_epochs}")
            self.model.train()
            loader = tqdm(self.data_manager.train_loader, desc=f"Train(e{epoch_idx})")
            for batch in loader:
                del batch["id"]
                batch = move_to_cuda_device(batch, self.config.device)
                result = self.model(**batch)
                loss = result["loss"]
                loader.set_postfix({"loss": loss.item()})
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            logger.info(loader)

            measures = self.eval("dev")
            self.history["dev"].append(measures)
            test_measures = self.eval("test")
            self.history["test"].append(test_measures)

            is_best = False
            if measures["f1"] > self.best_metric:
                is_best = True
                self.best_metric = measures["f1"]
                self.best_epoch = epoch_idx
                self.no_climbing_cnt = 0
            else:
                self.no_climbing_cnt += 1

            if is_best and self.config.save_best_ckpt:
                self.save_ckpt(epoch=epoch_idx)

            logger.info(
                f"Epoch: {epoch_idx}, is_best: {is_best}, Dev: {measures}, Test: {test_measures}"
            )

            if (
                self.config.num_early_stop > 0
                and self.no_climbing_cnt > self.config.num_early_stop
            ):
                break

        logger.info(
            (
                f"Trial finished. Best Epoch: {self.best_epoch}, Dev: {self.history['dev'][self.best_epoch]}, "
                f"Best Test: {self.history['test'][self.best_epoch]}"
            )
        )

    @torch.no_grad()
    def eval(self, dataset_name):
        self.model.eval()
        name2loader = {
            "train": self.data_manager.train_loader,
            "dev": self.data_manager.dev_loader,
            "test": self.data_manager.test_loader,
        }
        loader = tqdm(name2loader[dataset_name], desc=f"{dataset_name} Eval")
        preds = []
        golds = []
        for batch in loader:
            batch = move_to_cuda_device(batch, self.config.device)
            golds.extend(batch["labels"].detach().cpu().tolist())
            del batch["id"]
            del batch["labels"]
            result = self.model(**batch)
            preds.extend(result["pred"].long().detach().cpu().tolist())
        logger.info(loader)
        measures = mc_prf1(
            preds,
            golds,
            self.transform.label_encoder.num_tags,
            ignore_labels=[self.transform.label_encoder.label2id["NA"]],
        )
        return measures["macro"]

    @torch.no_grad()
    def predict(self, text, head, tail):
        self.model.eval()
        batch = self.transform.predict_transform(
            {"text": text, "head": head, "tail": tail}
        )
        tensor_batch = self.data_manager.collate_fn([batch])
        if "labels" in tensor_batch:
            del tensor_batch["labels"]
        if "id" in tensor_batch:
            del tensor_batch["id"]
        tensor_batch = move_to_cuda_device(tensor_batch, self.config.device)
        result = self.model(**tensor_batch)
        outs = result["pred"]
        preds = []
        for pred in outs:
            preds.append(self.transform.label_encoder.id2label[pred])

        return preds



from collections import defaultdict
from functools import reduce

from rex.metrics import calc_p_r_f1_from_tp_fp_fn


def tagging_prf1(gold_ents, pred_ents):
    # TODO: add macro and type-specified support
    measure_results = defaultdict(lambda: {"p": 0.0, "r": 0.0, "f1": 0.0})
    global_tp = global_fp = global_fn = 0
    for gold_ents, pred_ents in zip(gold_ents, pred_ents):
        gold_ents = set(reduce(lambda x, y: x + y, gold_ents.values(), []))
        pred_ents = set(reduce(lambda x, y: x + y, pred_ents.values(), []))
        intersection = gold_ents & pred_ents
        global_tp += len(intersection)
        global_fp += len(pred_ents - intersection)
        global_fn += len(gold_ents - intersection)

    measure_results["micro"] = calc_p_r_f1_from_tp_fp_fn(
        global_tp, global_fp, global_fn
    )
    return measure_results



from pathlib import Path

from omegaconf import OmegaConf
import torch
import torch.optim as optim
from rex.utils.config import ConfigParser
from rex.utils.logging import logger
from rex.utils.io import dump_json, load_line_json
from rex.data.dataset import CachedDataset
from rex.data.manager import CachedManager
from rex.utils.tensor_move import move_to_cuda_device
from rex.utils.progress_bar import tqdm
from rex.tasks.base_task import TaskBase
from rex.utils.initialization import init_all
from rex.utils.tagging import get_entities_from_tag_seq
from rex.metrics.tagging import tagging_prf1

from .data import CachedTaggingTransform, tagging_collate_fn
from .model import LSTMCRFModel


class NERTask(TaskBase):
    def __init__(self, config):
        super().__init__(config)

        self.middle_path = Path(config.task_dir).joinpath("middle")
        self.middle_path.mkdir(parents=True, exist_ok=True)

        self.transform = CachedTaggingTransform(config.max_seq_len)
        self.data_manager = CachedManager(
            config.train_filepath,
            config.dev_filepath,
            config.test_filepath,
            CachedDataset,
            self.transform,
            load_line_json,
            config.train_batch_size,
            config.eval_batch_size,
            tagging_collate_fn,
            debug_mode=config.debug_mode,
            load_train=not config.skip_train,
            load_dev=not config.skip_train,
            load_test=not config.skip_final_eval,
        )

        self.model = LSTMCRFModel(
            self.transform.vocab.size,
            config.emb_size,
            config.hidden_size,
            num_lstm_layers=config.num_lstm_layers,
            num_tags=self.transform.label_encoder.num_tags,
            dropout=config.dropout,
        )
        self.model = self.model.to(self.config.device)
        self.optimizer = optim.Adam(self.model.parameters(), lr=config.learning_rate)

    @torch.no_grad()
    def eval(self, dataset_name, verbose=False, postfix=""):
        self.model.eval()
        name2loader = {
            "train": self.data_manager.train_eval_loader,
            "dev": self.data_manager.dev_loader,
            "test": self.data_manager.test_loader,
        }
        loader = tqdm(
            name2loader[dataset_name], desc=f"{dataset_name} Eval", ncols=80, ascii=True
        )
        seqs = []
        tags = []
        preds = []
        for batch in loader:
            tags.extend(batch["tags"].detach().cpu().tolist())
            del batch["tags"]
            seqs.extend(batch["inputs"].detach().cpu().tolist())
            out = self.model(**batch)
            preds.extend(out)

        gold_ents_all = []
        pred_ents_all = []
        # num_illegal_tag = 0
        dump_data = []
        for seq, tag, pred in zip(seqs, tags, preds):
            seq = self.transform.vocab.convert_ids_to_tokens(seq)
            tag = self.transform.label_encoder.decode(tag)
            pred = self.transform.label_encoder.decode(pred)
            # num_illegal_tag += get_num_illegal_tags_from_tag_seq(pred)
            gold_ents = get_entities_from_tag_seq(seq, tag)
            pred_ents = get_entities_from_tag_seq(seq, pred)
            gold_ents_all.append(gold_ents)
            pred_ents_all.append(pred_ents)
            dump_data.append(
                {
                    "seq": seq,
                    "gold_tags": tag,
                    "pred_tags": pred,
                    "gold_ents": gold_ents,
                    "pred_ents": pred_ents,
                }
            )

        measures = tagging_prf1(gold_ents_all, pred_ents_all)

        if verbose:
            logger.info(f"{dataset_name} Eval Measures: {measures}")

        dump_json(
            dump_data, self.middle_path.joinpath(f"{dataset_name}.{postfix}.json")
        )

        return measures["micro"]["f1"]

    @torch.no_grad()
    def predict(self, string, verbose=False):
        self.model.eval()
        batch = self.transform.predict_transform(string)
        tensor_batch = self.data_manager.collate_fn([batch])
        if "labels" in tensor_batch:
            del tensor_batch["labels"]
        outs = self.model(**tensor_batch)
        pred_ents_per_sent = []
        pred_tag = outs["preds"][0]
        # in case of OOV
        refreshed_pred_ents = []
        pred_tag = self.transform.label_encoder.decode(pred_tag)
        pred_ents = get_entities_from_tag_seq(string, pred_tag)
        for ent_type in pred_ents:
            for ent in pred_ents[ent_type]:
                refreshed_pred_ents.append(
                    (string[ent[2] : ent[3]], ent[1], ent[2], ent[3])
                )
        pred_ents_per_sent.append(refreshed_pred_ents)

        if verbose:
            logger.info(f"Pred ents: {pred_ents_per_sent}")

        return pred_ents_per_sent

    def print_final_record(self):
        logger.info(
            (
                f"Best Epoch: {self.best_epoch}, Dev: {self.history['dev'][self.best_epoch]}, "
                f"Best Test: {self.history['test'][self.best_epoch]}"
            )
        )

    def train(self):
        for epoch_idx in range(self.config.num_epochs):
            logger.info(f"Epoch: {epoch_idx}/{self.config.num_epochs}")
            self.model.train()
            loader = tqdm(self.data_manager.train_loader, desc=f"Train(e{epoch_idx})")
            for batch in loader:
                batch = move_to_cuda_device(batch, self.config.device)
                result = self.model(**batch)
                loss = result["loss"]
                loader.set_postfix({"loss": loss.item()})
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            logger.info(loader)

            train_measures = self.eval("train", verbose=True, postfix=f"{epoch_idx}")
            self.history["train"].append(train_measures)
            dev_measures = self.eval("dev", verbose=True, postfix=f"{epoch_idx}")
            self.history["dev"].append(dev_measures)
            test_measures = self.eval("test", verbose=True, postfix=f"{epoch_idx}")
            self.history["test"].append(test_measures)

            measures = dev_measures
            is_best = False
            if measures > self.best_metric:
                is_best = True
                self.best_metric = measures
                self.best_epoch = epoch_idx
                self.no_climbing_cnt = 0
            else:
                self.no_climbing_cnt += 1

            if is_best and self.config.save_best_ckpt:
                self.save_ckpt(f"{epoch_idx}.{100 * dev_measures:.3f}", epoch_idx)
                self.save_ckpt("best", epoch_idx)

            logger.info(
                (
                    f"Epoch: {epoch_idx}, is_best: {is_best}, "
                    f"Train: {100 * train_measures:.3f}, "
                    f"Dev: {100 * dev_measures:.3f}, "
                    f"Test: {100 * test_measures:.3f}"
                )
            )

            if (
                self.config.num_early_stop > 0
                and self.no_climbing_cnt > self.config.num_early_stop
            ):
                break

        logger.info("Trial finished.")
        self.print_final_record()


def main():
    config = ConfigParser.parse_cmd()
    init_all(config.task_dir, config.random_seed, True, config)
    logger.info(OmegaConf.to_object(config))
    task = NERTask(config)
    logger.info(f"task: {type(task)}")

    if not config.skip_train:
        try:
            logger.info("Start Training")
            task.train()
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except Exception as err:
            logger.exception(err)
        finally:
            task.print_final_record()

    if not config.skip_final_eval:
        task.load(config.final_eval_model_filepath)
        task.eval("test", verbose=True)

    if not config.skip_predict_example:
        task.load(config.final_eval_model_filepath)
        text = "this is a sentence."
        task.predict(text, verbose=True)


if __name__ == "__main__":
    main()



import torch


def re_collate_fn(data):
    final_data = {
        "id": [],
        "token_ids": [],
        "mask": [],
        "labels": [],
        "head_pos": [],
        "tail_pos": [],
    }
    for d in data:
        for key in final_data:
            final_data[key].append(d[key])

    final_data["token_ids"] = torch.tensor(final_data["token_ids"], dtype=torch.long)
    final_data["mask"] = torch.tensor(final_data["mask"], dtype=torch.long)
    if all(x is not None for x in final_data["labels"]):
        final_data["labels"] = torch.tensor(final_data["labels"], dtype=torch.long)
    else:
        final_data["labels"] = None
    final_data["head_pos"] = torch.tensor(final_data["head_pos"], dtype=torch.long)
    final_data["tail_pos"] = torch.tensor(final_data["tail_pos"], dtype=torch.long)

    return final_data


def bag_re_collate_fn(data):
    final_data = {
        "id": [],
        "token_ids": [],
        "mask": [],
        "labels": [],
        "head_pos": [],
        "tail_pos": [],
        "scopes": [],
    }
    scope = []
    scope_begin = 0
    labels = []
    # data is a batch, for every bag in the batch:
    for bag in data:
        scope_ = [scope_begin]
        # for every instance in the bag
        for ins in bag:
            for key in final_data:
                if key != "scopes":
                    final_data[key].append(ins[key])
            scope_begin += 1
        labels.append(ins["labels"])
        scope_.append(scope_begin)
        scope.append(scope_)
    final_data["scopes"] = scope

    final_data["token_ids"] = torch.tensor(final_data["token_ids"], dtype=torch.long)
    final_data["mask"] = torch.tensor(final_data["mask"], dtype=torch.long)
    if all(x is not None for x in final_data["labels"]):
        final_data["labels"] = torch.tensor(labels, dtype=torch.long)
    else:
        final_data["labels"] = None
    final_data["head_pos"] = torch.tensor(final_data["head_pos"], dtype=torch.long)
    final_data["tail_pos"] = torch.tensor(final_data["tail_pos"], dtype=torch.long)

    return final_data


def subj_obj_span_collate_fn(data):
    final_data = {
        "id": [],
        "tokens": [],
        "entities": [],
        "relations": [],
        "token_ids": [],
        "mask": [],
        "subj_heads": [],
        "subj_tails": [],
        "one_subj": [],
        "subj2objs": [],
        "triples": [],
        "subj_head": [],
        "subj_tail": [],
        "obj_head": [],
        "obj_tail": [],
    }
    for d in data:
        for key in final_data:
            if key in d:
                final_data[key].append(d[key])

    final_data["token_ids"] = torch.tensor(final_data["token_ids"], dtype=torch.long)
    final_data["mask"] = torch.tensor(final_data["mask"], dtype=torch.long)
    for key in [
        "subj_heads",
        "subj_tails",
        "subj_head",
        "subj_tail",
        "obj_head",
        "obj_tail",
    ]:
        if final_data[key]:
            final_data[key] = torch.tensor(final_data[key], dtype=torch.float)

    return final_data



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from rex.modules.embeddings.static_embedding import StaticEmbedding
from rex.modules.cnn import MultiKernelCNN
from rex.modules.ffn import FFN
from transformers import AutoModel


class DummyModel(nn.Module):
    def __init__(
        self,
        vocab_size: Optional[int] = 5000,
        embedding_dim: Optional[int] = 300,
        in_channel: Optional[int] = 300,
        out_channel: Optional[int] = 300,
        num_classes: Optional[int] = 3,
        dropout: Optional[float] = 0.5,
    ):
        super().__init__()

        self.embedding = StaticEmbedding(vocab_size, embedding_dim, dropout=dropout)
        self.cnn = MultiKernelCNN(
            in_channel=in_channel,
            num_filters=out_channel,
            kernel_sizes=(1, 3, 5),
            dropout=dropout,
        )
        self.ffn = FFN(
            input_dim=out_channel,
            out_dim=num_classes,
            mid_dims=out_channel // 2,
            dropout=dropout,
            act_fn=nn.LeakyReLU(),
        )

        self.loss = nn.CrossEntropyLoss()

    def forward(self, input_ids, labels=None, **kwargs):
        embed = self.embedding(input_ids)
        hidden = self.cnn(embed)
        outs = self.ffn(hidden)

        probs, preds = outs.softmax(-1).max(-1)
        results = {"logits": outs, "probs": probs, "preds": preds}
        if labels is not None:
            results["loss"] = self.loss(outs, labels)
        return results


class DummyPLMModel(nn.Module):
    def __init__(
        self,
        plm_filepath: str,
        num_filters: Optional[int] = 300,
        num_classes: Optional[int] = 3,
        dropout: Optional[float] = 0.5,
    ):
        super().__init__()

        self.plm = AutoModel.from_pretrained(plm_filepath)
        hidden_size = self.plm.config.hidden_size
        self.cnn = MultiKernelCNN(
            in_channel=hidden_size,
            num_filters=num_filters,
            kernel_sizes=(1, 3, 5),
            dropout=dropout,
        )
        self.ffn = FFN(
            input_dim=hidden_size,
            out_dim=num_classes,
            mid_dims=hidden_size // 2,
            dropout=dropout,
            act_fn=nn.LeakyReLU(),
        )

        self.loss = nn.CrossEntropyLoss()

    def forward(self, input_ids, attention_mask, token_type_ids, labels=None, **kwargs):
        plm_outs = self.plm(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            return_dict=True,
        )
        hidden = self.cnn(plm_outs.last_hidden_state)
        outs = self.ffn(hidden)

        probs, preds = outs.softmax(-1).max(-1)
        results = {"logits": outs, "probs": probs, "preds": preds}
        if labels is not None:
            results["loss"] = self.loss(outs, labels)
        return results



from typing import Dict, Iterable, Any

from rex.data.vocab import Vocab


class TransformBase(object):
    def __init__(self, max_seq_len) -> None:
        super().__init__()
        self.max_seq_len = max_seq_len
        self.vocab = Vocab()

    def transform(self, lines: Iterable):
        raise NotImplementedError

    def predict_transform(self, strings: Iterable[str]):
        ret_data = []
        for string in strings:
            seq_len = min(len(string), self.max_seq_len)
            comp_len = max(0, (self.max_seq_len - seq_len))
            ret_data.append(
                {
                    "token_ids": self.vocab.convert_tokens_to_ids(list(string))[
                        :seq_len
                    ]
                    + comp_len * [self.vocab.pad_idx],
                    "mask": seq_len * [1] + comp_len * [0],
                }
            )
        return ret_data

    def __call__(self, *args: Any, **kwargs: Any):
        return self.transform(*args, **kwargs)



import os
import sys
import shutil
from pathlib import Path

import rex
from rex.utils.config import ConfigParser, ConfigArgument


REX_DIR = Path(rex.__file__).parent
TEMPLATES_DIR = REX_DIR.joinpath("templates")


def main():
    args = ConfigParser.parse_cmd(
        ConfigArgument("command", choices=["new"], help="REx mode."),
        init_priority_args=False,
        cmd_args=sys.argv[1:2],
    )
    if args.command == "new":
        new_args = ConfigParser.parse_cmd(
            ConfigArgument(
                "task_type",
                choices=["classification", "tagging"],
                help="Type of the new task",
            ),
            ConfigArgument(
                "task_name", help="Name of the task (used for creating new task folder)"
            ),
            init_priority_args=False,
            cmd_args=sys.argv[2:],
        )
        task_dir = Path(new_args.task_name)
        task_dir.mkdir(parents=True, exist_ok=False)
        src_dir = TEMPLATES_DIR.joinpath(new_args.task_type)
        for filename in os.listdir(src_dir):
            src_fp = src_dir.joinpath(filename)
            dest_fp = task_dir.joinpath(filename)
            shutil.copy2(src_fp, dest_fp)
        print(f"New {new_args.task_type} task in {task_dir.absolute()}")
    else:
        print(f"Unknown command: {args.command}")
        sys.exit(1)



import json
from pathlib import Path

from omegaconf import OmegaConf
import torch
import torch.optim as optim
from rex.utils.config import ConfigParser
from rex.utils.logging import logger
from rex.utils.io import dump_json, load_line_json
from rex.data.dataset import CachedDataset
from rex.data.manager import CachedManager
from rex.utils.tensor_move import move_to_cuda_device
from rex.utils.progress_bar import tqdm
from rex.tasks.base_task import TaskBase
from rex.utils.initialization import init_all
from rex.metrics.classification import mc_prf1

from .data import CachedMCTransform, mc_collate_fn
from .model import DummyPLMModel


class DummyTask(TaskBase):
    def __init__(self, config: OmegaConf) -> None:
        super().__init__(config)

        self.middle_path = Path(self.config.task_dir).joinpath("middle")
        self.middle_path.mkdir(parents=True, exist_ok=True)

        self.transform = CachedMCTransform(self.config.max_seq_len)
        self.data_manager = CachedManager(
            config.train_filepath,
            config.dev_filepath,
            config.test_filepath,
            CachedDataset,
            self.transform,
            load_line_json,
            config.train_batch_size,
            config.eval_batch_size,
            mc_collate_fn,
            debug_mode=config.debug_mode,
            load_train=not config.skip_train,
            load_dev=not config.skip_train,
            load_test=not config.skip_final_eval,
        )
        self.model = DummyPLMModel(
            config.plm_dir,
            config.num_filters,
            self.transform.label_encoder.num_classes,
            dropout=config.dropout,
        )
        self.model.to(self.config.device)
        self.optimizer = optim.Adam(self.model.parameters(), lr=config.learning_rate)

    def print_final_record(self):
        logger.info(
            (
                f"Best Epoch: {self.best_epoch}, Dev: {self.history['dev'][self.best_epoch]}, "
                f"Best Test: {self.history['test'][self.best_epoch]}"
            )
        )

    def train(self):
        for epoch_idx in range(self.config.num_epochs):
            logger.info(f"Epoch: {epoch_idx}/{self.config.num_epochs}")
            self.model.train()
            loader = tqdm(self.data_manager.train_loader, desc=f"Train(e{epoch_idx})")
            for batch in loader:
                batch = move_to_cuda_device(batch, self.config.device)
                result = self.model(**batch)
                loss = result["loss"]
                loader.set_postfix({"loss": loss.item()})
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            logger.info(loader)

            train_measures = self.eval("train", verbose=True, postfix=f"{epoch_idx}")
            self.history["train"].append(train_measures)
            dev_measures = self.eval("dev", verbose=True, postfix=f"{epoch_idx}")
            self.history["dev"].append(dev_measures)
            test_measures = self.eval("test", verbose=True, postfix=f"{epoch_idx}")
            self.history["test"].append(test_measures)

            measures = dev_measures
            is_best = False
            if measures > self.best_metric:
                is_best = True
                self.best_metric = measures
                self.best_epoch = epoch_idx
                self.no_climbing_cnt = 0
            else:
                self.no_climbing_cnt += 1

            if is_best and self.config.save_best_ckpt:
                self.save_ckpt(f"{epoch_idx}.{100 * dev_measures:.3f}", epoch_idx)
                self.save_ckpt("best", epoch_idx)

            logger.info(
                (
                    f"Epoch: {epoch_idx}, is_best: {is_best}, "
                    f"Train: {100 * train_measures:.3f}, "
                    f"Dev: {100 * dev_measures:.3f}, "
                    f"Test: {100 * test_measures:.3f}"
                )
            )

            if (
                self.config.num_early_stop > 0
                and self.no_climbing_cnt > self.config.num_early_stop
            ):
                break

        logger.info("Trial finished.")
        self.print_final_record()

    @torch.no_grad()
    def eval(self, dataset_name, verbose=False, postfix=""):
        self.model.eval()
        name2loader = {
            "train": self.data_manager.train_eval_loader,
            "dev": self.data_manager.dev_loader,
            "test": self.data_manager.test_loader,
        }
        loader = tqdm(name2loader[dataset_name], desc=f"{dataset_name} Eval")
        preds = []
        golds = []
        dump_data = []
        for raw_batch in loader:
            raw_batch = move_to_cuda_device(raw_batch, self.config.device)
            golds.extend(raw_batch["labels"])
            outputs = self.model(**raw_batch)
            batch_pred_ = outputs["preds"].detach().tolist()

            for id_, text, gold, pred_ in zip(
                raw_batch["id"],
                raw_batch["text"],
                raw_batch["labels"],
                batch_pred_,
            ):
                dump_data.append(
                    {
                        "id": id_,
                        "text": text,
                        "gold": gold,
                        "pred": pred_,
                    }
                )

        logger.info(loader)
        measures = mc_prf1(
            preds,
            golds,
            num_classes=self.transform.label_encoder.num_tags,
            label_idx2name=self.transform.label_encoder.id2label,
        )
        if verbose:
            logger.info(f"{dataset_name} Eval Measures: {measures}")

        dump_json(
            dump_data, self.middle_path.joinpath(f"{dataset_name}.{postfix}.json")
        )
        return measures["micro"]["f1"]

    @torch.no_grad()
    def predict(self, text, verbose=False):
        self.model.eval()
        data = {"id": "predict", "text": text}
        if verbose:
            logger.info(json.dumps(data, ensure_ascii=False, indent=2))

        preds = []

        batch = self.transform.predict_transform(data)
        tensor_batch = self.data_manager.collate_fn([batch])
        tensor_batch = move_to_cuda_device(tensor_batch, self.config.device)
        outputs = self.model(**tensor_batch)
        pred_label_id = outputs["preds"].cpu().tolist()[0]
        pred_label = self.transform.label_encoder.id2label[pred_label_id]

        if verbose:
            logger.info(f"{pred_label}")
        return preds


def main():
    config = ConfigParser.parse_cmd()
    init_all(config.task_dir, config.random_seed, True, config)
    logger.info(OmegaConf.to_object(config))
    task = DummyTask(config)
    logger.info(f"task: {type(task)}")

    if not config.skip_train:
        try:
            logger.info("Start Training")
            task.train()
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except Exception as err:
            logger.exception(err)
        finally:
            task.print_final_record()

    if not config.skip_final_eval:
        task.load(config.final_eval_model_filepath)
        task.eval("test", verbose=True)

    if not config.skip_predict_example:
        task.load(config.final_eval_model_filepath)
        text = "this is a sentence."
        task.predict(text, verbose=True)


if __name__ == "__main__":
    main()



import os
import logging
from pathlib import Path
from typing import Optional

import torch
from torch import distributed
from torch.nn import parallel
from omegaconf import OmegaConf

from rex.utils.logging import logger
from rex.utils.initialization import init_all
from rex.tasks import (
    CONFIG_PARAMS_FILENAME,
    CHECKPOINT_FILENAME_TEMPLATE,
    CHECKPOINT_DIRNAME,
    BEST_CHECKPOINT_FILENAME_TEMPLATE,
    BEST_IDENTIFIER,
)


class TaskBase(object):
    def __init__(
        self,
        config: OmegaConf,
        initialize: Optional[bool] = True,
        makedirs: Optional[bool] = True,
        dump_configfile: Optional[bool] = True,
    ) -> None:
        self.config = config
        self.model = None
        self.optimizer = None

        self.history = {"train": [], "dev": [], "test": []}
        self.no_climbing_cnt = 0
        self.best_metric = -100.0
        self.best_epoch = -1

        if initialize:
            init_all(
                config.task_dir, seed=config.random_seed, set_cudnn=True, config=None
            )

        config_string = OmegaConf.to_yaml(config, resolve=True)
        logger.info(config_string)

        if makedirs:
            if not os.path.exists(config.task_dir):
                os.makedirs(config.task_dir)
            else:
                logger.warning(f"Overwrite task dir: {config.task_dir}")

        if dump_configfile:
            OmegaConf.save(
                config,
                os.path.join(config.task_dir, CONFIG_PARAMS_FILENAME),
                resolve=True,
            )

    def train(self, *args, **kwargs):
        raise NotImplementedError

    def eval(self, *args, **kwargs):
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        raise NotImplementedError

    def load(
        self,
        path: str,
        load_model: Optional[bool] = True,
        load_optimizer: Optional[bool] = False,
    ):
        if os.path.exists(path):
            logger.info("Resume checkpoint from {}".format(path))
        else:
            raise ValueError("Checkpoint does not exist, {}".format(path))

        if torch.cuda.device_count() == 0:
            store_dict = torch.load(path, map_location="cpu")
        else:
            store_dict = torch.load(path, map_location=torch.device(self.config.device))

        self.config = OmegaConf.load(
            os.path.join(self.config.task_dir, CONFIG_PARAMS_FILENAME)
        )

        if load_model:
            if self.model and "model_state" in store_dict:
                if isinstance(self.model, parallel.DataParallel) or isinstance(
                    self.model, parallel.DistributedDataParallel
                ):
                    self.model.module.load_state_dict(store_dict["model_state"])
                else:
                    self.model.load_state_dict(store_dict["model_state"])
                logger.info("Load model successfully")
            else:
                raise ValueError(
                    f"Model loading failed. self.model={self.model}, stored_dict_keys={store_dict.keys()}"
                )
        else:
            logger.info("Not load model")

        if load_optimizer:
            if self.optimizer and "optimizer_state" in store_dict:
                self.optimizer.load_state_dict(store_dict["optimizer_state"])
                logger.info("Load optimizer successfully")
            else:
                raise ValueError(
                    f"Model loading failed. self.model={self.optimizer}, stored_dict_keys={store_dict.keys()}"
                )
        else:
            logger.info("Not load optimizer")

        self.history = store_dict.pop("history")
        self.no_climbing_cnt = store_dict.pop("no_climbing_cnt")
        self.best_metric = store_dict.pop("best_metric")
        self.best_epoch = store_dict.pop("best_epoch")

    def load_best_ckpt(self, load_optimizer: Optional[bool] = False):
        _task_dir = Path(self.config.task_dir)
        model_name = self.model.__class__.__name__
        ckpt_filepath = str(
            _task_dir.joinpath(
                CHECKPOINT_DIRNAME,
                BEST_CHECKPOINT_FILENAME_TEMPLATE.format(model_name),
            )
        )
        logger.info(f"Loading model from: {ckpt_filepath}")
        self.load(ckpt_filepath, load_model=True, load_optimizer=load_optimizer)

    def save(self, path, epoch: Optional[int] = None):
        logger.info(f"Dumping checkpoint into: {path}")
        store_dict = {}

        if self.model:
            if isinstance(self.model, parallel.DataParallel) or isinstance(
                self.model, parallel.DistributedDataParallel
            ):
                model_state = self.model.module.state_dict()
            else:
                model_state = self.model.state_dict()
            store_dict["model_state"] = model_state
        else:
            logger.info("No model state is dumped", logging.WARNING)

        if self.optimizer:
            store_dict["optimizer_state"] = self.optimizer.state_dict()
        else:
            logger.info("No optimizer state is dumped", logging.WARNING)

        if epoch:
            store_dict["epoch"] = epoch

        store_dict["history"] = self.history
        store_dict["no_climbing_cnt"] = self.no_climbing_cnt
        store_dict["best_metric"] = self.best_metric
        store_dict["best_epoch"] = self.best_epoch

        torch.save(store_dict, path)

    def save_ckpt(
        self, identifier: Optional[str] = BEST_IDENTIFIER, epoch: Optional[int] = None
    ):
        ckpt_dir = os.path.join(self.config.task_dir, CHECKPOINT_DIRNAME)
        if not os.path.exists(ckpt_dir):
            os.makedirs(ckpt_dir)

        ckpt_name = CHECKPOINT_FILENAME_TEMPLATE.format(
            self.model.__class__.__name__, identifier
        )
        self.save(os.path.join(ckpt_dir, ckpt_name), epoch)

    def logging(self, msg: str, level: Optional[int] = logging.INFO):
        if self.in_distributed_mode():
            msg = "Rank {} {}".format(distributed.get_rank(), msg)
        if self.config.only_master_logging:
            if self.is_master_node():
                logger.log(level, msg)
        else:
            logger.log(level, msg)

    def is_master_node(self):
        if self.in_distributed_mode():
            if distributed.get_rank() == 0:
                return True
            else:
                return False
        else:
            return True

    def in_distributed_mode(self):
        return self.config.local_rank >= 0

    @classmethod
    def from_configfile(
        cls,
        config_filepath: str,
        load_train_data: Optional[bool] = True,
        load_dev_data: Optional[bool] = True,
        load_test_data: Optional[bool] = True,
        **kwargs,
    ):
        logger.info(f"Initializing from configuration file: {config_filepath}")
        config = OmegaConf.load(config_filepath)

        return cls.from_config(
            config,
            load_train_data=load_train_data,
            load_dev_data=load_dev_data,
            load_test_data=load_test_data,
            **kwargs,
        )

    @classmethod
    def from_config(
        cls,
        config: OmegaConf,
        load_train_data: Optional[bool] = True,
        load_dev_data: Optional[bool] = True,
        load_test_data: Optional[bool] = True,
        **kwargs,
    ):
        logger.info(f"Initializing from configuration: {OmegaConf.to_yaml(config)}")

        # in case of any redundant memory taken when inference
        config["load_train_data"] = load_train_data
        config["load_dev_data"] = load_dev_data
        config["load_test_data"] = load_test_data

        kwargs["initialize"] = kwargs.pop("initialize", True)
        kwargs["makedirs"] = kwargs.pop("makedirs", True)
        kwargs["dump_configfile"] = kwargs.pop("dump_configfile", True)
        return cls(config, **kwargs)

    @classmethod
    def from_taskdir(
        cls,
        task_dir: str,
        load_best_model: Optional[bool] = True,
        load_best_optimizer: Optional[bool] = False,
        load_train_data: Optional[bool] = True,
        load_dev_data: Optional[bool] = True,
        load_test_data: Optional[bool] = True,
        **kwargs,
    ):
        _task_dir = Path(task_dir)
        config_filepath = str(_task_dir.joinpath(CONFIG_PARAMS_FILENAME).absolute())
        ins = cls.from_configfile(
            config_filepath,
            load_train_data=load_train_data,
            load_dev_data=load_dev_data,
            load_test_data=load_test_data,
            **kwargs,
        )
        if load_best_model:
            ins.load_best_ckpt(load_optimizer=load_best_optimizer)
        return ins



from typing import Callable, Optional

from torch.utils.data import DataLoader, Dataset


class CachedManager(object):
    def __init__(
        self,
        train_filepath: str,
        dev_filepath: str,
        test_filepath: str,
        dataset_class: Dataset,
        transform: Callable,
        load_fn: Callable,
        train_batch_size: int,
        eval_batch_size: int,
        collate_fn: Callable,
        train_shuffle: Optional[bool] = True,
        eval_shuffle: Optional[bool] = False,
        num_workers: Optional[int] = 4,
        debug_mode: Optional[bool] = False,
        load_train_data: Optional[bool] = True,
        load_dev_data: Optional[bool] = True,
        load_test_data: Optional[bool] = True,
    ):
        self.collate_fn = collate_fn
        self.train_set = None
        self.train_loader = None
        self.train_eval_loader = None
        self.dev_set = None
        self.dev_loader = None
        self.test_set = None
        self.test_loader = None

        if load_train_data:
            self.train_set = dataset_class(
                transform(load_fn(train_filepath), debug=debug_mode)
            )
            self.train_loader = DataLoader(
                self.train_set,
                batch_size=train_batch_size,
                shuffle=train_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )
            self.train_eval_loader = DataLoader(
                self.train_set,
                batch_size=eval_batch_size,
                shuffle=eval_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )

        if load_dev_data:
            self.dev_set = dataset_class(
                transform(load_fn(dev_filepath), debug=debug_mode)
            )
            self.dev_loader = DataLoader(
                self.dev_set,
                batch_size=eval_batch_size,
                shuffle=eval_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )

        if load_test_data:
            self.test_set = dataset_class(
                transform(load_fn(test_filepath), debug=debug_mode)
            )
            self.test_loader = DataLoader(
                self.test_set,
                batch_size=eval_batch_size,
                shuffle=eval_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )


class StreamTransformManager(object):
    def __init__(
        self,
        train_filepath: str,
        dev_filepath: str,
        test_filepath: str,
        dataset_class: Dataset,
        transform: Callable,
        load_fn: Callable,
        train_batch_size: int,
        eval_batch_size: int,
        collate_fn: Callable,
        train_shuffle: Optional[bool] = True,
        eval_shuffle: Optional[bool] = False,
        num_workers: Optional[int] = 4,
        debug_mode: Optional[bool] = False,
        load_train_data: Optional[bool] = True,
        load_dev_data: Optional[bool] = True,
        load_test_data: Optional[bool] = True,
    ):
        self.collate_fn = collate_fn

        if load_train_data:
            self.train_set = dataset_class(
                load_fn(train_filepath), transform, debug=debug_mode
            )
            self.train_loader = DataLoader(
                self.train_set,
                batch_size=train_batch_size,
                shuffle=train_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )
            self.train_eval_loader = DataLoader(
                self.train_set,
                batch_size=eval_batch_size,
                shuffle=eval_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )

        if load_dev_data:
            self.dev_set = dataset_class(
                load_fn(dev_filepath), transform, debug=debug_mode
            )
            self.dev_loader = DataLoader(
                self.dev_set,
                batch_size=eval_batch_size,
                shuffle=eval_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )

        if load_test_data:
            self.test_set = dataset_class(
                load_fn(test_filepath), transform, debug=debug_mode
            )
            self.test_loader = DataLoader(
                self.test_set,
                batch_size=eval_batch_size,
                shuffle=eval_shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )



import torch
import torch.nn as nn
from rex.modules.crf import PlainCRF


class LSTMCRFModel(nn.Module):
    def __init__(
        self, vocab_size, emb_size, hidden_size, num_lstm_layers, num_tags, dropout
    ):
        super().__init__()

        self.emb_enc = nn.Embedding(vocab_size, emb_size)
        self.lstm_enc = nn.LSTM(
            emb_size,
            hidden_size,
            num_layers=num_lstm_layers,
            bias=True,
            batch_first=True,
            dropout=dropout,
            bidirectional=True,
        )
        self.hidden2tag = nn.Linear(2 * hidden_size, num_tags)
        self.dropout = nn.Dropout(dropout)
        self.crf = PlainCRF(num_tags)

    def forward(self, input_ids, labels=None, mask=None, **kwargs):
        emb = self.emb_enc(input_ids)
        emb = self.dropout(emb)
        sorted_seq_lengths, indices = torch.sort(mask.sum(-1), descending=True)
        _, desorted_indices = torch.sort(indices, descending=False)
        emb = emb[indices]
        out = nn.utils.rnn.pack_padded_sequence(
            emb, sorted_seq_lengths.detach().cpu(), batch_first=True
        )
        out, (_, _) = self.lstm_enc(out)
        out, _ = nn.utils.rnn.pad_packed_sequence(
            out, batch_first=True, total_length=mask.size(-1)
        )
        out = out[desorted_indices]
        out = self.hidden2tag(out)

        results = {"preds": self.crf.decode(out)}
        if mask is not None:
            mask = mask.bool()
        if labels is not None:
            results["loss"] = -self.crf(out, labels, mask)

        return results
